"""Memo program constants."""
from solders.pubkey import Pubkey

MEMO_PROGRAM_ID: Pubkey = Pubkey.from_string("MemoSq4gqABAXKb96qnH8TysNcWxMyWCqXgDLGmfcHr")
"""Public key that identifies the Memo program."""
